UNIT GRAFPACK; {Borland Pascal 2-D & 3-D Graphics Library}

{1R2a- MC 1995 MAR 22 WED - separated mainstream from UNIT library functions}
{1R2 - MC 1990 APR 24 TUE - added MIDPOINT2D & MIDPOINT3D functions; SYSREAL}
{1R1 - MC 198? ??? ?? ??? - original version}

INTERFACE

USES
  SYSREALU;

CONST
  PIOVER2 = PI/2.000; {mathematical pi divided by 2}
  DEGRAD  = PI/180.0; {degrees-to-radians conversion factor}
  RADDEG  = 180.0/PI; {radians-to-degrees conversion factor}

TYPE
  AXIS    = (X,Y,Z);                 {Cartesian coordinate axis designator}
  POINT   = (A,B,C);                 {point designator}
  POINT2D = ARRAY [X..Y] OF SYSREAL; {Cartesian coordinate of a 2-D point}
  POINT3D = ARRAY [X..Z] OF SYSREAL; {Cartesian coordinate of a 3-D point}
  LINE2D  = ARRAY [A..B] OF POINT2D; {2 points to determine 2-D line}
  LINE3D  = ARRAY [A..B] OF POINT3D; {2 points to determine 3-D line}
  PLANE3D = ARRAY [A..C] OF POINT3D; {3 points to determine 3-D plane}
  TRAN2D  = ARRAY [1..3] OF ARRAY [1..3] OF SYSREAL; {2-D transform matrix}
  TRAN3D  = ARRAY [1..4] OF ARRAY [1..4] OF SYSREAL; {3-D transform matrix}
  STRING4   = STRING [4];
  STRING6   = STRING [6];
  STRING12  = STRING [12];
  STRING255 = STRING [255];

VAR
  AXR         : SYSREAL; {global floating point registers}
  I,J,K,L,M,N : INTEGER; {global indexing registers}

FUNCTION RSGN (REALNUMBER : SYSREAL) : SYSREAL;
FUNCTION ISGN (INTEGRALNUMBER : INTEGER) : INTEGER;
FUNCTION DECIMALFEET (FEET : INTEGER; INCHES : INTEGER) : SYSREAL;
FUNCTION FEETINCHES (DECIMALFEET : SYSREAL) : STRING12;
FUNCTION DEGREES (RADS : SYSREAL) : SYSREAL;
FUNCTION RADIANS (DEGR : SYSREAL) : SYSREAL;
FUNCTION ANGLEPOINT2DORIGIN (POINT : POINT2D) : SYSREAL;
FUNCTION DISTANCE2D (A,B : POINT2D) : SYSREAL;
FUNCTION DISTANCE3D (A,B : POINT3D) : SYSREAL;

PROCEDURE ZMAT3D (VAR TMAT : TRAN3D);
PROCEDURE TRANSLATE2D (VAR SOURCE : POINT2D; DESTINATION : POINT2D);
PROCEDURE TRANSLATE3D (VAR SOURCE : POINT3D; DESTINATION : POINT3D);
PROCEDURE ISWAP (VAR A,B : INTEGER);
PROCEDURE RSWAP (VAR A,B : SYSREAL);
PROCEDURE SSWAP (VAR A,B : STRING255);
PROCEDURE ATRANSLATE3D (VAR TMAT : TRAN3D; NEWORIGIN : POINT3D);
PROCEDURE SCALE3D (VAR TMAT : TRAN3D; SCALINGFACTOR : POINT3D);
PROCEDURE ROTATE3D (VAR TMAT : TRAN3D; ROTANGLE : SYSREAL; RAXIS : AXIS);
PROCEDURE MULTIPLY4X4 (A,B : TRAN3D; VAR C : TRAN3D);
PROCEDURE T3D (APOINT : POINT3D; TMAT : TRAN3D; VAR T3DP : POINT3D);
PROCEDURE GEN3DTRANSFORM (NEWORIGIN,ZOOMFACTOR,TILT : POINT3D; VAR TMAT : TRAN3D);
PROCEDURE CMULT4X4 (VAR TMAT : TRAN3D; CONSTANT : SYSREAL);
PROCEDURE GEN3DOBSERVE (VAR TMAT : TRAN3D; VANTAGEPOINT : POINT3D);
PROCEDURE MIDPOINT2D (A,B : POINT2D; VAR C : POINT2D);
PROCEDURE MIDPOINT3D (A,B : POINT3D; VAR C : POINT3D);

IMPLEMENTATION

USES
  CRT,TRIGPACK;

FUNCTION RSGN (REALNUMBER : SYSREAL) : SYSREAL;
 {This function returns +1.0 if REALNUMBER is positive, -1.0 if
  REALNUMBER is negative, or 0.0 if REALNUMBER is zero.}
BEGIN
  IF REALNUMBER = 0.0
    THEN RSGN := 0.0
    ELSE IF REALNUMBER > 0.0
      THEN RSGN := +1.0
      ELSE RSGN := -1.0;
END;

FUNCTION ISGN (INTEGRALNUMBER : INTEGER) : INTEGER;
 {This function returns +1 if INTEGRALNUMBER is positive, -1 if
  INTEGRALNUMBER is negative, or 0 if INTEGRALNUMBER is zero.}
BEGIN
  IF INTEGRALNUMBER = 0
    THEN ISGN := 0
    ELSE IF INTEGRALNUMBER > 0
      THEN ISGN := +1
      ELSE ISGN := -1;
END;

PROCEDURE ZMAT3D (VAR TMAT : TRAN3D);
 {This procedure zeroes out the 3-D transform matrix which is passed.}
VAR
  I, J : INTEGER;
BEGIN
  FOR I := 1 TO 4 DO
    FOR J := 1 TO 4 DO
      TMAT [I,J] := 0.0;
END;

FUNCTION DECIMALFEET (FEET : INTEGER; INCHES : INTEGER) : SYSREAL;
 {This function converts feet and inches into decimal feet.}
BEGIN
  DECIMALFEET := FEET + (INCHES / 12);
END;

FUNCTION FEETINCHES (DECIMALFEET : SYSREAL) : STRING12;
 {This function converts decimal feet into a string of the format
  9999'99.999", where the 'Z's indicate zero-suppressed decimal positions
  and the '9's indicate zero-show decimal positions.  The foot (') and
  inch (") symbol appear literally, as does the decimal point (.).}
VAR
  FSTRING : STRING4;
  ISTRING : STRING6;
  RSTRING : STRING12;
BEGIN
  STR ((INT(DECIMALFEET)):4:0,FSTRING);
  STR (((DECIMALFEET - INT(DECIMALFEET)) * 12):6:3,ISTRING);
  RSTRING := FSTRING + '''' + ISTRING + '"';
  FOR K := 1 TO LENGTH (RSTRING) DO BEGIN
    IF (RSTRING [K] = ' ') THEN RSTRING [K] := '0';
  END;
  FEETINCHES := RSTRING;
END;

PROCEDURE TRANSLATE2D (VAR SOURCE : POINT2D; DESTINATION : POINT2D);
 {This procedure translates the 2-D point SOURCE to a 2-D space in which
  the origin is located at the point DESTINATION, where DESTINATION's
  location is specified in coordinates of SOURCE's 2-D space.}
BEGIN
  SOURCE [X] := SOURCE [X] - DESTINATION [X];
  SOURCE [Y] := SOURCE [Y] - DESTINATION [Y];
END;

PROCEDURE TRANSLATE3D (VAR SOURCE : POINT3D; DESTINATION : POINT3D);
 {This procedure translates the 3-D point SOURCE to a 3-D space in which
  the origin is located at the point DESTINATION, where DESTINATION's
  location is specified in coordinates of SOURCE's 3-D space.}
BEGIN
  SOURCE [X] := SOURCE [X] - DESTINATION [X];
  SOURCE [Y] := SOURCE [Y] - DESTINATION [Y];
  SOURCE [Z] := SOURCE [Z] - DESTINATION [Z];
END;

FUNCTION DEGREES (RADS : SYSREAL) : SYSREAL;
 {This procedure converts degrees to radians.}
BEGIN
  DEGREES := RADS * RADDEG;
END;

FUNCTION RADIANS (DEGR : SYSREAL) : SYSREAL;
 {This function converts radians to degrees.}
BEGIN
  RADIANS := DEGR * DEGRAD;
END;

PROCEDURE ISWAP (VAR A,B : INTEGER);
 {This procedure swaps the 2 integers passed to it.}
VAR
  C : INTEGER;
BEGIN
  C := A;
  A := B;
  B := C;
END;

PROCEDURE RSWAP (VAR A,B : SYSREAL);
 {This procedure swaps the 2 real numbers passed to it.}
VAR
  C : SYSREAL;
BEGIN
  C := A;
  A := B;
  B := C;
END;

PROCEDURE SSWAP (VAR A,B : STRING255);
 {This procedure swaps the 2 strings passed to it.}
VAR
  C : STRING255;
BEGIN
  C := A;
  A := B;
  B := C;
END;

PROCEDURE ATRANSLATE3D (VAR TMAT : TRAN3D; NEWORIGIN : POINT3D);
 {This procedure produces a transformation matrix which will translate
  points from the old 3-D space to the new 3-D space which has its ori-
  gin at point NEWORIGIN expressed in terms of the old 3-D space's co-
  ordinates.}
VAR
  I, J : INTEGER;
BEGIN
  ZMAT3D (TMAT);
  FOR I := 1 TO 4 DO TMAT [I,J] := 1.0;
  TMAT [1,4] := -NEWORIGIN [X];
  TMAT [2,4] := -NEWORIGIN [Y];
  TMAT [3,4] := -NEWORIGIN [Z];
END;

PROCEDURE SCALE3D (VAR TMAT : TRAN3D; SCALINGFACTOR : POINT3D);
 {This procedure produces a 3-D transformation matrix which can be used
  to transform all points in a 3-D space.  Each axis can have a unique
  scaling factor, so the scaling factors for all 3 axes must be speci-
  fied.  This is done by giving the factors in 3-D point form, where it
  is understood that the X-coordinate is in effect the X-axis scaling
  factor.}
VAR
  I, J : INTEGER;
BEGIN
  ZMAT3D (TMAT);
  TMAT [1,1] := SCALINGFACTOR [X];
  TMAT [2,2] := SCALINGFACTOR [Y];
  TMAT [3,3] := SCALINGFACTOR [Z];
  TMAT [4,4] := 1.0;
END;

PROCEDURE ROTATE3D (VAR TMAT : TRAN3D; ROTANGLE : SYSREAL; RAXIS : AXIS);
 {This procedure produces a 3-D transform matrix which can be used to
  rotate all points about the axis specified by RAXIS.}
VAR
  AXNUM, POS1, POS2 : INTEGER;
  ANGLESIN, ANGLECOS : SYSREAL;
BEGIN
  CASE RAXIS OF
    X : AXNUM := 1;
    Y : AXNUM := 2;
    Z : AXNUM := 3;
    ELSE HALT
  END;
  TMAT [4,4] := 0.0;
  TMAT [AXNUM,AXNUM] := 0.0;
  POS1 := (AXNUM MOD 3) + 1;
  POS2 := (POS1 MOD 3) + 1;
  ANGLESIN := SIN (ROTANGLE);
  ANGLECOS := COS (ROTANGLE);
  TMAT [POS1,POS1] := ANGLECOS;
  TMAT [POS2,POS2] := ANGLECOS;
  TMAT [POS1,POS2] := ANGLESIN;
  TMAT [POS2,POS1] := -ANGLESIN;
END;

PROCEDURE MULTIPLY4X4 (A,B : TRAN3D; VAR C : TRAN3D);
 {This procedure multiplies 2 - 4x4 matrices giving a 4x4 matrix as the result.}
VAR
  I, J, K : INTEGER;
  S : SYSREAL;
BEGIN
  ZMAT3D (C);
  FOR I := 1 TO 4 DO BEGIN
    FOR J := 1 TO 4 DO BEGIN
      S := 0.0;
      FOR K := 1 TO 4 DO S := S + A[I,K] * B[K,J];
      C[I,J] := S;
    END;
  END;
END;

PROCEDURE T3D (APOINT : POINT3D; TMAT : TRAN3D; VAR T3DP : POINT3D);
 {This procedure returns the coordinates of a 3-D point which has been
  transformed from the original space based upon the transformation
  matrix TMAT.  This is done simply by multiplying:  APOINT x TMAT.}
BEGIN
  T3DP [X] := (TMAT [1,1] * APOINT [X]) + (TMAT [1,2] * APOINT [Y])
   + (TMAT [1,3] * APOINT [Z]) + (TMAT [1,4]);
  T3DP [Y] := (TMAT [2,1] * APOINT [X]) + (TMAT [2,2] * APOINT [Y])
   + (TMAT [2,3] * APOINT [Z]) + (TMAT [2,4]);
  T3DP [Z] := (TMAT [3,1] * APOINT [X]) + (TMAT [3,2] * APOINT [Y])
   + (TMAT [3,3] * APOINT [Z]) + (TMAT [3,4]);
END;

PROCEDURE GEN3DTRANSFORM (NEWORIGIN,ZOOMFACTOR,TILT : POINT3D; VAR TMAT : TRAN3D);
 {This procedure generates a 3-D transformation matrix which translates
  the point to a 3-D space with origin at NEWORIGIN, compresses and/or
  expands along the X, Y, and Z axes according to the compansion factors
  specified by ZOOMFACTOR, and rotates the object along the X, Y, and Z
  axes by the tilt factors specified by TILT.}
VAR
  A, B, C, D, E, F, G, H : TRAN3D;
BEGIN
  ATRANSLATE3D (A,NEWORIGIN);
  SCALE3D (B,ZOOMFACTOR);
  ROTATE3D (C,TILT [X],X);
  ROTATE3D (D,TILT [Y],Y);
  ROTATE3D (E,TILT [Z],Z);
  MULTIPLY4X4 (B,A,F);
  MULTIPLY4X4 (C,F,G);
  MULTIPLY4X4 (D,G,H);
  MULTIPLY4X4 (E,H,TMAT);
END;

PROCEDURE CMULT4X4 (VAR TMAT : TRAN3D; CONSTANT : SYSREAL);
 {This procedure multiplies a 4x4 matrix by a constant.}
VAR
  I, J : INTEGER;

BEGIN
  FOR I := 1 TO 4 DO BEGIN
    FOR I := 1 TO 4 DO BEGIN
      TMAT [I,J] := TMAT [I,J] * CONSTANT;
    END;
  END;
END;

FUNCTION ANGLEPOINT2DORIGIN (POINT : POINT2D) : SYSREAL;
 {This function returns the angle in radians between a 2-D point and
  the origin with respect to the X-axis.}
VAR
  ANGLE : SYSREAL;
BEGIN
  IF (ABS (POINT [X]) < 0.00001)
    THEN ANGLEPOINT2DORIGIN := PI / 2.0
    ELSE
      IF (POINT [Y] < 0.00001)
        THEN ANGLEPOINT2DORIGIN := 0.0
        ELSE BEGIN
          ANGLE := ARCTAN (POINT [Y] / POINT [X]);
          IF (POINT [X] < 0.0) THEN ANGLE := ANGLE + PI;
          ANGLEPOINT2DORIGIN := ANGLE;
        END;
END;

PROCEDURE GEN3DOBSERVE (VAR TMAT : TRAN3D; VANTAGEPOINT : POINT3D);
 {This procedure generates a 4x4 matrix which is used to transform
  3-D points to appear as though they are being observed from the
  specified VANTAGEPOINT.}
VAR
  THETA, PHI, PSI, RADIUS : SYSREAL;
  VP2D                    : POINT2D;
  A,B,C,D,E,F,G,H,I,J,K   : TRAN3D;
BEGIN
  VP2D [X] := VANTAGEPOINT [X];
  VP2D [Y] := VANTAGEPOINT [Y];
  THETA := ANGLEPOINT2DORIGIN (VP2D);
  ROTATE3D (A,THETA,X);
  RADIUS := SQRT(SQR(VANTAGEPOINT [X]) + SQR(VANTAGEPOINT [Y]));
  VP2D [X] := VANTAGEPOINT [Z];
  VP2D [Y] := RADIUS;
  PHI := ANGLEPOINT2DORIGIN (VP2D);
  PHI := PHI + PI;
  ROTATE3D (B,PHI,Y);
  MULTIPLY4X4 (B,A,C);
  VP2D [X] := C [2,2];
  VP2D [Y] := C [1,2];
  PSI := ANGLEPOINT2DORIGIN (VP2D);
  ROTATE3D (C,PSI,Z);
  MULTIPLY4X4 (C,TMAT,D);
END;

FUNCTION DISTANCE2D (A,B : POINT2D) : SYSREAL;
 {This function returns the distance between 2 points in 2-D space.}
BEGIN
  DISTANCE2D := SQRT(SQR(A[X]-B[X]) + SQR(A[Y]-B[Y]));
END;

FUNCTION DISTANCE3D (A,B : POINT3D) : SYSREAL;
 {This function returns the distance between 2 points in 3-D space.}
BEGIN
  DISTANCE3D := SQRT(SQR(A[X]-B[X]) + SQR(A[Y]-B[Y]) + SQR(A[Z]-B[Z]));
END;

PROCEDURE MIDPOINT2D (A,B : POINT2D; VAR C : POINT2D);
 {This computes the midpoint between points A & B in 2-D space.}
BEGIN
  C [X] := (A[X]+B[X])/2.0;
  C [Y] := (A[Y]+B[Y])/2.0;
END;

PROCEDURE MIDPOINT3D (A,B : POINT3D; VAR C : POINT3D);
 {This computes the midpoint between points A & B in 3-D space.}
BEGIN
  C [X] := (A[X]+B[X])/2.0;
  C [Y] := (A[Y]+B[Y])/2.0;
  C [Z] := (A[Z]+B[Z])/2.0;
END;

END.
