UNIT TRIGPACK; {Trigonometric & Hyperbolic Function Library}

{2R1 - MC 1995 MAR 22 WED - converted to Borland Pascal 7.0; public distribution}
{1R2 - MC 1988 NOV 06 SAT - converted to Turbo Pascal 5.0}
{1R1 - MC 1988 APR 11 MON - converted to Turbo Pascal 4.0}
{0R1 - MC 1984 ??? ?? ??? - original version}

INTERFACE

USES
  SYSREALU;

FUNCTION RADIANS (DEGS : SYSREAL) : SYSREAL;
FUNCTION DEGREES (RADS : SYSREAL) : SYSREAL;

{FUNCTION SIN (THETA : SYSREAL) : SYSREAL;} {internal BP function}
{FUNCTION COS (THETA : SYSREAL) : SYSREAL;} {internal BP function}
FUNCTION TAN (THETA : SYSREAL) : SYSREAL;
FUNCTION COT (THETA : SYSREAL) : SYSREAL;
FUNCTION SEC (THETA : SYSREAL) : SYSREAL;
FUNCTION CSC (THETA : SYSREAL) : SYSREAL;

FUNCTION ARCSIN (THETASIN : SYSREAL) : SYSREAL;
FUNCTION ARCCOS (THETACOS : SYSREAL) : SYSREAL;
{FUNCTION ARCTAN (THETATAN : SYSREAL) : SYSREAL;} {internal BP function}
FUNCTION ARCCOT (THETACOT : SYSREAL) : SYSREAL;
FUNCTION ARCSEC (THETASEC : SYSREAL) : SYSREAL;
FUNCTION ARCCSC (THETACSC : SYSREAL) : SYSREAL;

FUNCTION SINH (THETA : SYSREAL) : SYSREAL;
FUNCTION COSH (THETA : SYSREAL) : SYSREAL;
FUNCTION TANH (THETA : SYSREAL) : SYSREAL;
FUNCTION COTH (THETA : SYSREAL) : SYSREAL;
FUNCTION SECH (THETA : SYSREAL) : SYSREAL;
FUNCTION CSCH (THETA : SYSREAL) : SYSREAL;

FUNCTION ARCSINH (THETASINH : SYSREAL) : SYSREAL;
FUNCTION ARCCOSH (THETACOSH : SYSREAL) : SYSREAL;
FUNCTION ARCTANH (THETATANH : SYSREAL) : SYSREAL;
FUNCTION ARCCOTH (THETACOTH : SYSREAL) : SYSREAL;
FUNCTION ARCSECH (THETASECH : SYSREAL) : SYSREAL;
FUNCTION ARCCSCH (THETACSCH : SYSREAL) : SYSREAL;

IMPLEMENTATION

USES
  CRT;

TYPE
  ERRORSTRING = STRING [64]; {function error message}

CONST
  PIOVER2    = PI/2.0;   {pre-computed PI factors}
  PIOVER180  = PI/180.0;
  I180OVERPI = 180.0/PI;
  ERRFACT    = 1.0E-003; {error factor of 1/10th of 1%}

PROCEDURE TRIGERROR (ERRORREASON : ERRORSTRING);
 {This procedure is invoked by any function which detects an arithmetic
  error (e.g. ARCSIN(32)).  A message is displayed stating that TRIGPACK has
  detected an error along with the actual reason.  The program is then halted.}
BEGIN
  TEXTMODE (LASTMODE);
  WRITELN ('*TRIGPACK* Error: ',ERRORREASON);
  HALT (1);
END;

FUNCTION RADIANS (DEGS : SYSREAL) : SYSREAL;
 {N.B. Parentheses are not used so compiler can generate optimum code,
  which for 80x87 chips means not having to reload 10 byte registers.}
BEGIN
  RADIANS := DEGS*PIOVER180;
END;

FUNCTION DEGREES (RADS : SYSREAL) : SYSREAL;
 {N.B. Parentheses are not used so compiler can generate optimum code,
  which for 80x87 chips means not having to reload 10 byte registers.}
BEGIN
  DEGREES := RADS*I180OVERPI;
END;

FUNCTION RSGN (REALNUMBER : SYSREAL) : SYSREAL;
 {This function returns +1.0 if REALNUMBER is positive, -1.0 if
  REALNUMBER is negative, or 0.0 if REALNUMBER is zero.}
BEGIN
  IF (REALNUMBER = 0.0)
    THEN RSGN := 0.0
    ELSE IF (REALNUMBER > 0.0)
      THEN RSGN := +1.0
      ELSE RSGN := -1.0;
END;

FUNCTION ISGN (INTEGRALNUMBER : INTEGER) : INTEGER;
 {This function returns +1 if INTEGRALNUMBER is positive, -1 if
  INTEGRALNUMBER is negative, or 0 if INTEGRALNUMBER is zero.}
BEGIN
  IF (INTEGRALNUMBER = 0)
    THEN ISGN := 0
    ELSE IF (INTEGRALNUMBER > 0)
      THEN ISGN := +1
      ELSE ISGN := -1;
END;

FUNCTION TAN (THETA : SYSREAL) : SYSREAL;
 {This function returns the tangent of radian angle THETA.}
BEGIN
  IF (COS(THETA) = 0.0)
    THEN TAN := MAXREAL
    ELSE TAN := SIN(THETA)/COS(THETA);
END;

FUNCTION COT (THETA : SYSREAL) : SYSREAL;
 {This function returns the cotangent of radian angle THETA.}
BEGIN
  IF (SIN(THETA) = 0.0)
    THEN COT := MAXREAL
    ELSE COT := COS(THETA)/SIN(THETA);
END;

FUNCTION SEC (THETA : SYSREAL) : SYSREAL;
 {This function returns the secant of radian angle THETA.}
BEGIN
  IF (COS(THETA) = 0.0)
    THEN SEC := MAXREAL
    ELSE SEC := 1.0/COS(THETA);
END;

FUNCTION CSC (THETA : SYSREAL) : SYSREAL;
 {This function returns the cosecant of radian angle THETA.}
BEGIN
  IF (SIN(THETA) = 0.0)
    THEN CSC := MAXREAL
    ELSE CSC := 1.0/SIN(THETA);
END;

FUNCTION ARCSIN (THETASIN : SYSREAL) : SYSREAL;
 {This function returns the arc sine of THETASIN.}
BEGIN
  IF (ABS(THETASIN) = 1.0)
    THEN ARCSIN := RSGN(THETASIN)*PIOVER2
    ELSE ARCSIN := ARCTAN(THETASIN/SQRT(-SQR(THETASIN)+1.0));
END;

FUNCTION ARCCOS (THETACOS : SYSREAL) : SYSREAL;
 {This function returns the arc cosine of THETACOS.}
BEGIN
  IF (THETACOS = 1.0)
    THEN ARCCOS := 0.0
    ELSE IF (THETACOS = -1.0)
      THEN ARCCOS := PI
      ELSE ARCCOS := PIOVER2-ARCTAN(THETACOS/SQRT(-SQR(THETACOS)+1.0));
END;

FUNCTION ARCCOT (THETACOT : SYSREAL) : SYSREAL;
 {This function returns the arc cotangent of THETACOT.}
BEGIN
  ARCCOT := PIOVER2-ARCTAN(THETACOT);
END;

FUNCTION ARCSEC (THETASEC : SYSREAL) : SYSREAL;
 {This function returns the arc secant of THETASEC.}
BEGIN
  IF (THETASEC > -1.0) AND (THETASEC < 1.0)
    THEN TRIGERROR ('ARCSEC(X) with -1 < X < +1');
  IF (THETASEC = -1.0)
    THEN ARCSEC := PI
    ELSE IF (THETASEC = 1.0)
      THEN ARCSEC := 0.0
      ELSE ARCSEC := PIOVER2-ARCTAN((1.0/THETASEC)/SQRT(-SQR((1.0/THETASEC))+1.0));
END;

FUNCTION ARCCSC (THETACSC : SYSREAL) : SYSREAL;
 {This function returns the arc cosecant of THETACSC.}
BEGIN
  IF (THETACSC > -1.0) AND (THETACSC < 1.0)
    THEN TRIGERROR ('ARCCSC(X) with -1 < X < +1');
  IF (THETACSC = -1.0)
    THEN ARCCSC := -PIOVER2
    ELSE IF (THETACSC = 1.0)
      THEN ARCCSC := PIOVER2
      ELSE ARCCSC := ARCTAN((1.0/THETACSC)/SQRT(-SQR((1.0/THETACSC))+1.0));
END;

FUNCTION SINH (THETA : SYSREAL) : SYSREAL;
 {This function returns the hyperbolic sine of THETA.}
BEGIN
  SINH := (EXP(THETA)-EXP(-THETA))/2.0;
END;

FUNCTION COSH (THETA : SYSREAL) : SYSREAL;
 {This function returns the hyperbolic cosine of THETA.}
BEGIN
  COSH := (EXP(THETA)+EXP(-THETA))/2.0;
END;

FUNCTION TANH (THETA : SYSREAL) : SYSREAL;
 {This function returns the hyperbolic tangent of THETA.}
VAR
  EXPPOS : SYSREAL;
  EXPNEG : SYSREAL;
BEGIN
  EXPPOS := EXP(THETA);
  EXPNEG := EXP(-THETA);
  TANH := (EXPPOS-EXPNEG)/(EXPPOS+EXPNEG);
END;

FUNCTION COTH (THETA : SYSREAL) : SYSREAL;
 {This function returns the hyperbolic cotangent of THETA.}
VAR
  EXPPOS : SYSREAL;
  EXPNEG : SYSREAL;
BEGIN
  IF (THETA = 0.0)
    THEN COTH := MAXREAL
    ELSE BEGIN
      EXPPOS := EXP(THETA);
      EXPNEG := EXP(-THETA);
      COTH := (EXPPOS+EXPNEG)/(EXPPOS-EXPNEG);
    END;
END;

FUNCTION SECH (THETA : SYSREAL) : SYSREAL;
 {This function returns the hyperbolic secant of THETA.}
BEGIN
  SECH := 2.0/(EXP(THETA)+EXP(-THETA));
END;

FUNCTION CSCH (THETA : SYSREAL) : SYSREAL;
 {This function returns the hyperbolic cosecant of THETA.}
BEGIN
  IF (THETA = 0.0)
    THEN CSCH := MAXREAL
    ELSE CSCH := 2.0/(EXP(THETA)-EXP(-THETA));
END;

FUNCTION ARCSINH (THETASINH : SYSREAL) : SYSREAL;
 {This function returns the arc hyperbolic sine of THETASINH.}
BEGIN
  ARCSINH := LN (THETASINH+SQRT(SQR(THETASINH)+1.0));
END;

FUNCTION ARCCOSH (THETACOSH : SYSREAL) : SYSREAL;
 {This function returns the arc hyperbolic cosine of THETACOSH.}
BEGIN
  IF (THETACOSH < 1.0)
    THEN TRIGERROR ('ARCCOSH(X) with X < 1')
    ELSE ARCCOSH := LN(THETACOSH+SQRT(SQR(THETACOSH)-1.0));
END;

FUNCTION ARCTANH (THETATANH : SYSREAL) : SYSREAL;
 {This function returns the arc hyperbolic tangent of THETATANH.}
BEGIN
  IF (ABS(THETATANH) > 1.0)
    THEN TRIGERROR ('ARCTANH(X) with ABS(X) > 1');
  IF (ABS(THETATANH) = 1.0)
    THEN ARCTANH := MAXREAL
    ELSE ARCTANH := 0.5*LN((1.0+THETATANH)/(1.0-THETATANH));
END;

FUNCTION ARCCOTH (THETACOTH : SYSREAL) : SYSREAL;
 {This function returns the arc hyperbolic cotangent of THETACOTH.}
BEGIN
  IF (ABS(THETACOTH) < 1.0)
    THEN TRIGERROR ('ARCCOTH(X) with ABS(X) < 1');
  IF (ABS(THETACOTH) = 1.0)
    THEN ARCCOTH := MAXREAL
    ELSE ARCCOTH := 0.5*LN((THETACOTH+1.0)/(THETACOTH-1.0));
END;

FUNCTION ARCSECH (THETASECH : SYSREAL) : SYSREAL;
 {This function returns the arc hyperbolic secant of THETASECH.}
BEGIN
  IF (THETASECH > 1.0)
    THEN TRIGERROR ('ARCSECH(X) with X > 1');
  IF (THETASECH < 0.0)
    THEN TRIGERROR ('ARCSECH(X) with X < 0');
  IF (THETASECH = 0.0)
    THEN ARCSECH := MAXREAL
    ELSE ARCSECH := LN((1.0/THETASECH)+SQRT((1.0/SQR(THETASECH))-1.0));
END;

FUNCTION ARCCSCH (THETACSCH : SYSREAL) : SYSREAL;
 {This function returns the arc hyperbolic cosecant of THETACSCH.}
BEGIN
  IF (ABS(THETACSCH) < ERRFACT)
    THEN ARCCSCH := MAXREAL
    ELSE IF (ABS(THETACSCH) < MAXREAL)
      THEN ARCCSCH := LN((1.0/THETACSCH)+SQRT((1.0/SQR(THETACSCH))+1.0))
      ELSE ARCCSCH := 0.0;
END;

END.
